/* Copyright (c) 2003 The Nutch Organization.  All rights reserved.   */
/* Use subject to the conditions in http://www.nutch.org/LICENSE.txt. */

package net.nutch.db;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

import net.nutch.io.*;
import net.nutch.pagedb.*;
import net.nutch.linkdb.*;

/***********************************************
 * The DistributedWebDBWriter creates a complicated 
 * distributed webdb that can be updated by many processors
 * on many spindles simultaneously.
 *
 * @author Mike Cafarella
 ***********************************************/

/********************************************************
 * The SegmentDescription class tells how many segments we have.
 *
 * We need an easy way of describing the segmentation of our
 * four databases.  That is, we need to know where to break
 * the database apart.
 *
 * To do so, we have hard-coded values that should (roughly)
 * segment each database's keyspace into 32 equal parts.
 * If we want fewer parts, we just merge neighboring segments.
 * But we cannot split these segments - 32 is the most we
 * will have.
 *
 * We also indicate the number of interesting bits for each
 * database type.  The values that describe the boundaries
 * must partition the region from 0 .. max-value-of-number-in-numBytes.  
 * "numBytes" cannot be any larger than 8.
 *
 * The items that belong in segment i are all the keys whose
 * first "numBytes" are less than or equal to the value at 
 * position i in the array indicated here.
 *
 * If the interesting part of the key doesn't happen right
 * at the start, we can make use of the "offset" value.  We
 * compute the key's segment by starting to read at that position,
 * and going into the key by "bits" positions.
 *
 * We should derive these values from looking at the likely
 * distribution of keys.  For keys based on MD5 hash values,
 * we trust we'll get an even distribution and our job is easy!  
 * But for keys based on URL-text, we need to actually examine 
 * the numbers.  
 *
 * The SegmentDescription class encapsulates the above logic,
 * and is instantiated statically for each of our database types.
 ********************************************************/
class SegmentDescription {
    int offset;
    int numBytes;
    long dividers[];

    /**
     * Create a seg desc out of the bit-offset, the num of bits,
     * and the set of dividing points
     */
    public SegmentDescription(int offset, int numBytes, long dividers[]) {
        this.offset = offset;
        this.numBytes = numBytes;
        this.dividers = dividers;
        if (numBytes > 8) {
            throw new RuntimeException("numBytes is >8: " + numBytes);
        }
    }

    /**
     * Compute the correct segment, given an input key bitpattern
     */
    public int determineSegment(byte key[]) {
        // First, extract the key region that we want
        // to pay attention to.
        long inputKey = 0;
        for (int i = offset; (i < numBytes + offset) && (i < key.length); i++) {
            byte b = key[i];
            inputKey = inputKey << 8;
            inputKey |= (0x00FF & key[i]);
        }

        // Now, find the position in the dividers array
        // such that the input key is less than or equal
        // to the value at the position.
        for (int i = 0; i < dividers.length; i++) {
            if (inputKey <= dividers[i] || ((i + 1) == dividers.length)) {
                return i;
            }
        }

        // Something's gone horribly wrong!
        throw new RuntimeException("Could not find segment for input key");
    }
}


/*****************************************
 * The DistributedWebDBWriter offers the usual
 * function calls of an IWebDBWriter.  However, there's
 * an additional necessary step.  After all the calls have
 * been made, there's an "apply" step, which takes as input
 * which "subsegment" should be processed.  This is the
 * step that allows us to execute in parallel.
 *
 *****************************************/
public class DistributedWebDBWriter implements IWebDBWriter {
    //
    // To choose which segment a data item belongs to, we need
    // to use a SegmentDescription.  Each such SD needs a
    // set of 32 values that divide its keyspace into equal parts.
    //

    //
    // Some URLs are more likely than others, so we choose the
    // "dividing values" empirically.  These numbers are derived
    // from examining the DMOZ URL set.
    //
    static long[] url8ByteDividers = {0x64657672616e646fl,
                                      0x686f6178696e682el,
                                      0x6d6167796e2e636fl,
                                      0x6f7368612d656e66l,
                                      0x74656d706c65656dl,
                                      0x7769656d2e6f6e65l,
                                      0x7777772e6165726fl,
                                      0x7777772e616e7468l,
                                      0x7777772e6261726fl,
                                      0x7777772e62727563l,
                                      0x7777772e6368656el,
                                      0x7777772e636f7268l,
                                      0x7777772e64697665l,
                                      0x7777772e656e7669l,
                                      0x7777772e666f7274l,
                                      0x7777772e67656f63l,
                                      0x7777772e67726565l,
                                      0x7777772e68746d69l,
                                      0x7777772e6a313033l,
                                      0x7777772e6c616b65l,
                                      0x7777772e6d617363l,
                                      0x7777772e6d726d70l,
                                      0x7777772e6e6a6472l,
                                      0x7777772e7062732el,
                                      0x7777772e71756565l,
                                      0x7777772e73616261l,
                                      0x7777772e736b6565l,
                                      0x7777772e73757267l,
                                      0x7777772e746f6e79l,
                                      0x7777772e75766d2el,
                                      0x7777772e77696767l,
                                      0x7FFFFFFFFFFFFFFFl};
    //
    // For the MD5, we expect an even distribution.  We can divide
    // the 1-byte space into 32 even parts.
    //
    static long[] digest1ByteDividers = {0x8, 0x10, 0x18, 0x20, 0x28, 0x30, 0x38, 0x40, 0x48, 0x50, 0x58, 0x60, 0x68, 0x70, 0x78, 0x80, 0x88, 0x90, 0x98, 0xa0, 0xa8, 0xb0, 0xb8, 0xc0, 0xc8, 0xd0, 0xd8, 0xe0, 0xe8, 0xf0, 0xf8, 0xFF};

    //
    // Create static SegmentDescription objs out of those dividing regions.
    //
    static SegmentDescription urlSegments = new SegmentDescription(7, 8, url8ByteDividers);
    static SegmentDescription md5Segments = new SegmentDescription(0, 1, digest1ByteDividers);

    static int NUM_SEGMENTS = 32;
    static int PAGEDB_TABLE = 0;
    static int PAGEDB_INDEX = 1;
    static int LINKDB_TABLE = 2;
    static int LINKDB_INDEX = 3;

    static String PAGEDB_TABLE_FILENAME = "pagedbtable.updates";
    static String PAGEDB_INDEX_FILENAME = "pagedbindex.updates";
    static String LINKDB_TABLE_FILENAME = "linkdbtable.updates";
    static String LINKDB_INDEX_FILENAME = "linkdbindex.updates";

    static int ADD_PAGE = 0;
    static int ADD_PAGE_IFN_PRESENT = 1;
    static int DEL_PAGE = 2;
    static int ADD_LINK = 3;

    /**
     * Make a brand-new Distributed Web DB on disk.
     */
    public static void createDistributedWebDB(File dir, int pageCacheMb, int linkCacheMb) throws IOException {
        DistributedWebDBWriter starter = new DistributedWebDBWriter(dir, pageCacheMb, linkCacheMb);
        starter.close();
    }

    File dir, updatesDir;
    DataOutputStream pageDbTable[], pageDbIndex[], linkDbTable[], linkDbIndex[];
    /**
     * Create a brand-new distributed webdb writer on disk
     */
    public DistributedWebDBWriter(File dir, int pageCacheMb, int linkCacheMb) throws IOException {
        this.dir = dir;
    }

    /**
     */
    public synchronized void close() throws IOException {
        closeUpdates();
    }

    /**
     */
    public void sync() throws IOException {
    }

    /**
     */
    public synchronized void addPage(Page page) throws IOException {
        DataOutputStream out = determineOutputStream(PAGEDB_TABLE, page.getURL().getBytes());
        out.writeInt(ADD_PAGE);
        page.write(out);
    }

    public synchronized void addPageWithScore(Page page) throws IOException {
        DataOutputStream out = determineOutputStream(PAGEDB_TABLE, page.getURL().getBytes());
        out.writeInt(ADD_PAGE);
        page.write(out);
    }

    /**
     */
    public synchronized void addPageIfNotPresent(Page page, Link link) throws IOException {
        addPageIfNotPresent(page);
    }

    /**
     */
    public synchronized void addPageIfNotPresent(Page page) throws IOException {
        DataOutputStream out = determineOutputStream(PAGEDB_TABLE, page.getURL().getBytes());
        out.writeInt(ADD_PAGE_IFN_PRESENT);
        page.write(out);
    }

    // Obsolete!
    public Enumeration pages() {
        return null;
    }

    /**
     */
    public synchronized void deletePage(String url) throws IOException {
        DataOutputStream out = determineOutputStream(PAGEDB_TABLE, url.getBytes());
        out.writeInt(DEL_PAGE);
        out.writeUTF(url);

        // Deleting is a serial operation.  This instruction
        // will generate a 2nd.
    }

    /**
     */
    public synchronized void addLink(Link lr) throws IOException {
        DataOutputStream out = determineOutputStream(LINKDB_TABLE, lr.getURL().getBytes());
        out.writeInt(ADD_LINK);
        lr.write(out);

        out = determineOutputStream(LINKDB_INDEX, lr.getFromID().getDigest());
        out.writeInt(ADD_LINK);
        lr.write(out);
    }

    /**
     * Apply changes for a given segment.  This is the stage-two
     * operation.
     */
    public synchronized void applyChanges(int segmentId) throws IOException {
        closeUpdates();

        Vector changeDirs = new Vector();
        File updates = new File(dir, "updates");
        String entries[] = updates.list();
        for (int i = 0; i < entries.length; i++) {
            File curSet = new File(updates, entries[i]);
            if (curSet.isDirectory()) {
                String dirName = (segmentId < 10) ? ("update.0" + segmentId) : ("update." + segmentId);
                changeDirs.add(new File(curSet, dirName));
            }
        }

        //PageDB pagedb = new PageDB(new File(dir, "pagedb"), 10, false, true);
        //LinkDB linkdb = new LinkDB(new File(dir, "linkdb"), 10, false, true);
        try {
            // Here are all the update-directories we're interested in.
            for (Enumeration e = changeDirs.elements(); e.hasMoreElements(); ) {
                File dir = (File) e.nextElement();
                File pagedbTableUpdates = new File(dir, PAGEDB_TABLE_FILENAME);
                File pagedbIndexUpdates = new File(dir, PAGEDB_INDEX_FILENAME);
                File linkdbTableUpdates = new File(dir, LINKDB_TABLE_FILENAME);
                File linkdbIndexUpdates = new File(dir, LINKDB_INDEX_FILENAME);

                DataOutputStream out = null;

                // Process page table
                if (pagedbTableUpdates.exists()) {
                    DataInputStream in = new DataInputStream(new FileInputStream(pagedbTableUpdates));
                    try {
                        while (in.available() > 0) {
                            int op = in.readInt();
                            if (op == ADD_PAGE) {
                                Page p = Page.read(in);
                                Page oldP = null;
                                //Page oldP = pagedb.setRecordToTable(p);

                                // Processed the table-insert.  Now must
                                // process index-remove and index-insert
                                if (oldP != null && ! p.getMD5().equals(oldP.getMD5())) {
                                    out = determineOutputStream(PAGEDB_INDEX, oldP.getMD5().getDigest());
                                    out.writeInt(DEL_PAGE);
                                    oldP.write(out);
                                }

                                out = determineOutputStream(PAGEDB_INDEX, p.getMD5().getDigest());
                                out.writeInt(ADD_PAGE);
                                p.write(out);
                            } else if (op == ADD_PAGE_IFN_PRESENT) {
                                Page p = Page.read(in);
                                /**
                                if (pagedb.getRecord(p.getURL()) == null) {
                                    Page oldP = pagedb.setRecordToTable(p);

                                    if (oldP != null && ! p.getMD5().equals(oldP.getMD5())) {
                                        out = determineOutputStream(PAGEDB_INDEX, oldP.getMD5().getDigest());
                                        out.writeInt(DEL_PAGE);
                                        oldP.write(out);
                                    }

                                    out = determineOutputStream(PAGEDB_INDEX, p.getMD5().getDigest());
                                    out.writeInt(ADD_PAGE);
                                    p.write(out);
                                }
                                **/
                            } else if (op == DEL_PAGE) {
                                UTF8 url = new UTF8();
                                url.readFields(in);
                                /**
                                Page p = pagedb.deleteFromTable(url);
                                out = determineOutputStream(PAGEDB_INDEX, p.getMD5().getDigest());
                                out.writeInt(DEL_PAGE);
                                p.write(out);
                                **/
                            }
                        }
                    } finally {
                        in.close();
                    }
                }

                // Process page index
                if (pagedbIndexUpdates.exists()) {
                    DataInputStream in = new DataInputStream(new FileInputStream(pagedbIndexUpdates));
                    try {
                        while (in.available() > 0) {
                            int op = in.readInt();
                            if (op == ADD_PAGE) {
                                // We pass in null here because we already
                                // deleted the old page via a DEL_PAGE
                                // command.  We should change the PageDB
                                // API so that this is not necessary.
                                // REMIND - mjc
                                //pagedb.setRecordToIndex(Page.read(in), null);
                            } else if (op == DEL_PAGE) {
                                //pagedb.deleteFromIndex(Page.read(in));
                            }
                        }
                    } finally {
                        in.close();
                    }
                }

                // Process link table
                if (linkdbTableUpdates.exists()) {
                    DataInputStream in = new DataInputStream(new FileInputStream(linkdbTableUpdates));
                    try {
                        while (in.available() > 0) {
                            int op = in.readInt();
                            if (op == ADD_LINK) {
                                //linkdb.setRecordToTable(LinkRecord.read(in));
                            }
                        }
                    } finally {
                        in.close();
                    }
                }

                // Process link index
                if (linkdbIndexUpdates.exists()) {
                    DataInputStream in = new DataInputStream(new FileInputStream(linkdbIndexUpdates));
                    try {
                        while (in.available() > 0) {
                            int op = in.readInt();
                            if (op == ADD_LINK) {
                                //linkdb.setRecordToIndex(LinkRecord.read(in));
                            }
                        }
                    } finally {
                        in.close();
                    }
                }

                pagedbTableUpdates.delete();
                pagedbIndexUpdates.delete();
                linkdbTableUpdates.delete();
                linkdbIndexUpdates.delete();
                dir.delete();
            }
        } finally {
            //pagedb.close();
            //linkdb.close();
        }
    }

    /**
     */
    private void closeUpdates() throws IOException {
        if (updatesDir != null) {
            for (int i = 0; i < NUM_SEGMENTS; i++) {
                pageDbTable[i].close();
                pageDbIndex[i].close();
                linkDbTable[i].close();
                linkDbIndex[i].close();
            }
            updatesDir = null;
        }
    }

    /**
     */
    private void needUpdates() throws IOException {
        if (updatesDir == null) {
            File updates = new File(dir, "updates");
            SimpleDateFormat formatter = new SimpleDateFormat("yyyyMMdd_HHmmss");

            // Build this processor's working updates directory, 
            // which we'll need if we need to add updates.
            while (true) {
                StringBuffer name = new StringBuffer();
                Date d = new Date(System.currentTimeMillis());
                formatter.format(d, name, new FieldPosition(0));
                try {
                    name.append("_" + InetAddress.getLocalHost().getCanonicalHostName());
                } catch (UnknownHostException uhe) {
                }

                this.updatesDir = new File(updates, name.toString());
                if (! this.updatesDir.exists()) {
                    this.updatesDir.mkdirs();
                    break;
                } else {
                    try {
                        Thread.sleep(1500);
                    } catch (InterruptedException ie) {
                    }
                }
            }

            // Create subdirs if necessary
            for (int i = 0; i < NUM_SEGMENTS; i++) {
                String name = (i < 10) ? ("update.0" + i) : ("update." + i);
                File curSegment = new File(updatesDir, name);
                if (! curSegment.exists()) {
                    curSegment.mkdir();
                }
            }

            // Create open streams to the segments.
            pageDbTable = new DataOutputStream[NUM_SEGMENTS];
            pageDbIndex = new DataOutputStream[NUM_SEGMENTS];
            linkDbTable = new DataOutputStream[NUM_SEGMENTS];
            linkDbIndex = new DataOutputStream[NUM_SEGMENTS];
            for (int i = 0; i < NUM_SEGMENTS; i++) {
                String name = (i < 10) ? ("update.0" + i) : ("update." + i);
                File curSegment = new File(updatesDir, name);
                try {
                    pageDbTable[i] = new DataOutputStream(new FileOutputStream(new File(curSegment, PAGEDB_TABLE_FILENAME)));
                    pageDbIndex[i] = new DataOutputStream(new FileOutputStream(new File(curSegment, PAGEDB_INDEX_FILENAME)));
                    linkDbTable[i] = new DataOutputStream(new FileOutputStream(new File(curSegment, LINKDB_TABLE_FILENAME)));
                    linkDbIndex[i] = new DataOutputStream(new FileOutputStream(new File(curSegment, LINKDB_INDEX_FILENAME)));
                } catch (Exception ie) {
                    ie.printStackTrace(System.out);
                }
            }
        }
    }

    /**
     */
    private DataOutputStream determineOutputStream(int table, byte key[]) throws IOException {
        needUpdates();
        if (table == PAGEDB_TABLE) {
            int segment = urlSegments.determineSegment(key);
            return pageDbTable[segment];
        } else if (table == PAGEDB_INDEX) {
            int segment = md5Segments.determineSegment(key);
            return pageDbIndex[segment];
        } else if (table == LINKDB_TABLE) {
            int segment = urlSegments.determineSegment(key);
            return linkDbTable[segment];
        } else if (table == LINKDB_INDEX) {
            int segment = md5Segments.determineSegment(key);
            return linkDbIndex[segment];
        }
        return null;
    }
    
    /**
     *
     */
    public static void main(String argv[]) throws IOException {
        if (argv.length < 2) {
            System.out.println("java net.nutch.db.DistributedWebDBWriter <webdb> [-create] [-makechanges <dmozfile>] [-applychanges <processorId>] [-addrandom <numadds>]");
            return;
        }

        String command = null, dmozfile = null;
        int processorId = -1, numAdds = 0;
        String db = argv[0];
        for (int i = 1; i < argv.length; i++) {
            if ("-create".equals(argv[i])) {
                command = argv[i];
            } else if ("-makechanges".equals(argv[i])) {
                command = argv[i];
                dmozfile = argv[i+1];
                i++;
            } else if ("-applychanges".equals(argv[i])) {
                command = argv[i];
                processorId = Integer.parseInt(argv[i+1]);
                i++;
            } else if ("-addrandom".equals(argv[i])) {
                command = argv[i];
                numAdds = Integer.parseInt(argv[i+1]);
                i++;
            }
        }

        if ("-create".equals(command)) {
            DistributedWebDBWriter.createDistributedWebDB(new File(argv[0]), 1, 1);
        } else if ("-addrandom".equals(command)) {
            Random rand = new Random();
            DistributedWebDBWriter writer = new DistributedWebDBWriter(new File(db), 1, 1);
            for (int i = 0; i < numAdds; i++) {
                StringBuffer buf = new StringBuffer();
                buf.append("http://");
                for (int j = 0; j < 20; j++) {
                    buf.append((char) (65 + Math.abs(rand.nextInt() % 26)));
                }
                String randUrl = buf.toString();
                writer.addPage(new Page(randUrl, 0.2f, 0.2f));
            }
            writer.close();
        } else if ("-makechanges".equals(command)) {
            DistributedWebDBWriter writer = new DistributedWebDBWriter(new File(db), 1, 1);
            WebDBInjector injector = new WebDBInjector(writer);
            try {
                injector.injectDmozFile(new File(dmozfile), 500, false, true, 0);
            } catch (org.xml.sax.SAXException se) {
                se.printStackTrace(System.out);
            } catch (javax.xml.parsers.ParserConfigurationException pce) {                
                pce.printStackTrace(System.out);
            }
            injector.close();
        } else if ("-applychanges".equals(command)) {
            DistributedWebDBWriter writer = new DistributedWebDBWriter(new File(db), 1, 1);
            writer.applyChanges(processorId);
            writer.close();
        }
    }
}


