/* Copyright (c) 2003 The Nutch Organization.  All rights reserved.   */
/* Use subject to the conditions in http://www.nutch.org/LICENSE.txt. */

package net.nutch.fetcher;

import net.nutch.pagedb.FetchListEntry;
import net.nutch.net.protocols.http.Http;
import net.nutch.net.protocols.http.Response;
import net.nutch.io.*;
import net.nutch.db.*;
import net.nutch.util.*;

import java.io.*;
import java.net.*;
import java.util.*;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.logging.Handler;

import org.cyberneko.html.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import org.apache.html.dom.*;

/* A simple fetcher. */
public class Fetcher { 
  public static final Logger LOG =
    LogFormatter.getLogger("net.nutch.fetcher.Fetcher");

  private ArrayFile.Reader fetchList;              // the input

  private ArrayFile.Writer fetcherDb;              // the output
  private ArrayFile.Writer rawDb;
  private ArrayFile.Writer strippedDb;

  private int threadCount = 30;                   // max number of threads

  private long start;                             // start time of fetcher run

  private long bytes;                             // total bytes fetched
  private int pages;                              // total pages fetched
  private int errors;                             // total pages errored

  private ThreadGroup group = new ThreadGroup("fetcher"); // our thread group

  private Http http = new Http();

  /** A fetcher thread. */
  private class FetcherThread extends Thread {
    private DOMFragmentParser parser = new DOMFragmentParser();

    public FetcherThread() {
      super(group, "starting");
    }

    public void run() {
      while (true) {
        FetchListEntry fle = null;
        String urlString = null;
        try {
          setName("starting");
          fle = (FetchListEntry)fetchList.next(new FetchListEntry());
          if (fle == null) return;

          urlString = fle.getPage().getURL().toString();
          setName(urlString);

          if (!fle.getFetch()) {                  // don't fetch this page
            LOG.fine("not fetching " + urlString);
            handleNoFetch(fle, true);
            continue;
          }

          URL url = new URL(urlString);           // parse the URL

          Response response = http.getResponse(url); // get HTTP response

          handleFetch(url, fle, response);

          synchronized (Fetcher.this) {           // record success
            pages++;
            bytes += response.getContent().length;
            if ((pages % 100) == 0)               // show status every 100pp
              status();
          }
        } catch (Throwable t) {
          if (fle != null) {
            LOG.info("fetch of " + urlString + " failed with: " + t);
            synchronized (Fetcher.this) {          // record failure
              errors++;
            }
            handleNoFetch(fle, false);
          }
        }
      }
    }

    private void handleFetch(URL url, FetchListEntry fle,
                             Response response)
      throws IOException, SAXException {
      String contentType = response.getHeader("Content-Type");
      if (contentType != null && !contentType.startsWith("text/html"))
        throw new IOException("Unknown content-type: " + contentType);
      
      DocumentFragment node = new HTMLDocumentImpl().createDocumentFragment();
      parser.parse(new InputSource            // parse content
                   (new ByteArrayInputStream(response.getContent())),
                   node);
      
      StringBuffer sb = new StringBuffer();
      getText(sb, node);
      String text = sb.toString();
      
      sb.setLength(0);
      getTitle(sb, node);
      String title = sb.toString().trim();
      
      ArrayList l = new ArrayList();
      getOutlinks(url, l, node);
      Outlink[] outlinks = (Outlink[])l.toArray(new Outlink[l.size()]);
      LOG.fine("found " + outlinks.length + " outlinks in " + url);

      outputPage(new FetcherOutput(fle, MD5Hash.digest(response.getContent()),
                                   true, title, outlinks),
                 new FetcherContent(response.getContent()),
                 new FetcherText(text));
    }

    private void handleNoFetch(FetchListEntry fle, boolean success) {
      outputPage(new FetcherOutput(fle, MD5Hash.digest(fle.getPage().getURL().toString()),
                                   success, "", new Outlink[0]),
                 new FetcherContent(new byte[0]),
                 new FetcherText(""));
    }
  }
      


  private void outputPage(FetcherOutput fo, FetcherContent raw,
                          FetcherText stripped) {
    try {
      synchronized (fetcherDb) {
        fetcherDb.append(fo);
        rawDb.append(raw);
        strippedDb.append(stripped);
      }
    } catch (Throwable t) {
      LOG.severe("error writing output:" + t.toString());
    }
  }
                                       

  private static void getText(StringBuffer sb, Node node) {
    if (node.getNodeType() == Node.TEXT_NODE) {
      sb.append(node.getNodeValue());             //.append("\n");
    }
    NodeList children = node.getChildNodes();
    if ( children != null ) {
      int len = children.getLength();
      for ( int i = 0; i < len; i++ ) {
        getText(sb, children.item(i));
      }
    }
  }

  private static boolean getTitle(StringBuffer sb, Node node) {
    if (node.getNodeType() == Node.ELEMENT_NODE) {
      if ("title".equalsIgnoreCase(node.getNodeName())) {
        getText(sb, node);
        return true;
      }
    }
    NodeList children = node.getChildNodes();
    if (children != null) {
      int len = children.getLength();
      for (int i = 0; i < len; i++) {
        if (getTitle(sb, children.item(i))) {
          return true;
        }
      }
    }
    return false;
  }

  private static void getOutlinks(URL base, ArrayList outlinks, Node node) {
    if (node.getNodeType() == Node.ELEMENT_NODE) {
      if ("a".equalsIgnoreCase(node.getNodeName())) {
        StringBuffer linkText = new StringBuffer();
        getText(linkText, node);

        NamedNodeMap attrs = node.getAttributes();
        String target= null;
        for (int i= 0; i < attrs.getLength(); i++ ) {
          if ("href".equalsIgnoreCase(attrs.item(i).getNodeName())) {
            target= attrs.item(i).getNodeValue();
            break;
          }
        }
        if (target != null)
          try {
            URL url = new URL(base, target);
            outlinks.add(new Outlink(url.toString(),
                                     linkText.toString().trim()));
          } catch (MalformedURLException e) {
            // don't care
          }
      }
    }
    NodeList children = node.getChildNodes();
    if ( children != null ) {
      int len = children.getLength();
      for ( int i = 0; i < len; i++ ) {
        getOutlinks(base, outlinks, children.item(i));
      }
    }
  }	
			
  /** Constructs a fetcher. */
  public Fetcher(String directory) throws IOException {
    fetchList = new ArrayFile.Reader
      (new File(directory, FetchListEntry.DIR_NAME).toString());
    fetcherDb = new ArrayFile.Writer
      (new File(directory, FetcherOutput.DIR_NAME).toString(), FetcherOutput.class);
    rawDb = new ArrayFile.Writer
      (new File(directory, FetcherContent.DIR_NAME).toString(), FetcherContent.class);
    strippedDb = new ArrayFile.Writer
      (new File(directory, FetcherText.DIR_NAME).toString(), FetcherText.class);
  }

  /** Set the thread count. */
  public void setThreadCount(int threadCount) {this.threadCount=threadCount;}

  /** Return the Http implementation. */
  public Http getHttp() { return http;}

  /** Set the logging level. */
  public void setLogLevel(Level level) {
    LOG.setLevel(level);
    Http.LOG.setLevel(level);
    LOG.info("logging at " + level);
  }

  /** Runs the fetcher. */
  public void run() throws IOException, InterruptedException {
    start = System.currentTimeMillis();
    for (int i = 0; i < threadCount; i++) {       // spawn threads
      FetcherThread thread = new FetcherThread(); 
      thread.start();
    }
    do {
      Thread.sleep(1000);
    } while (group.activeCount() > 0);            // wait for threads to finish

    fetchList.close();                            // close databases
    fetcherDb.close();
    rawDb.close();
    strippedDb.close();

    status();                                     // print final status
  }

  /** Display the status of the fetcher run. */
  public synchronized void status() {
    long ms = System.currentTimeMillis() - start;
    LOG.info("status: "
             + pages + " pages, "
             + errors + " errors, "
             + bytes + " bytes, "
             + ms + " ms");
    LOG.info("status: "
             + (((float)pages)/(ms/1000.0f))+" pages/s, "
             + (((float)bytes*8/1024)/(ms/1000.0f))+" kb/s, "
             + (((float)bytes)/pages) + " bytes/page");
  }

  /** Run the fetcher. */
  public static void main(String[] args) throws Exception {
    int timeout = -1;
    int threadCount = -1;
    boolean verbose = false;
    String directory = null;

    String usage = "Usage: Fetcher [-verbose] [-timeout N] [-threads M] dir";

    if (args.length == 0) {
      System.err.println(usage);
      System.exit(-1);
    }
      
    for (int i = 0; i < args.length; i++) {       // parse command line
      if (args[i].equals("-timeout")) {		  // found -timeout option
        timeout = Integer.parseInt(args[++i]) * 1000;
      } else if (args[i].equals("-threads")) {	  // found -threads option
        threadCount =  Integer.parseInt(args[++i]);
      } else if (args[i].equals("-verbose")) {	  // found -verbose option
        verbose = true;
      } else if (i != args.length-1) {
        System.err.println(usage);
        System.exit(-1);
      } else                                      // root is required parameter
        directory = args[i];
    }

    Fetcher fetcher = new Fetcher(directory);     // make a Fetcher
    if (timeout != -1)                            // set timeout option
      fetcher.getHttp().setTimeout(timeout);
    if (threadCount != -1)                        // set threadCount option
      fetcher.setThreadCount(threadCount);
                                                  // set log level
    fetcher.setLogLevel(verbose ? Level.FINE : Level.INFO);

    fetcher.run();                                // run the Fetcher
  }
}
