/* Copyright (c) 2003 The Nutch Organization.  All rights reserved.   */
/* Use subject to the conditions in http://www.nutch.org/LICENSE.txt. */

package net.nutch.indexer;

import net.nutch.pagedb.*;
import net.nutch.linkdb.*;
import net.nutch.fetcher.*;
import net.nutch.analysis.NutchDocumentAnalyzer;
import net.nutch.db.*;
import net.nutch.io.*;
import net.nutch.util.*;
import java.util.*;
import java.io.*;
import cgl.library.*;


import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;

import java.util.logging.Logger;
import java.util.Date;
import java.io.File;
import java.io.EOFException;

/** Creates an index for the output corresponding to a single fetcher run. */
public class IndexSegment {
  public static final String DONE_NAME = "index.done";
  public static final Logger LOG =
    LogFormatter.getLogger("net.nutch.index.IndexSegment");

  private float scorePower = NutchConf.getFloat("indexer.score.power", 0.5f);

  private int maxTitleLength =
    NutchConf.getInt("indexer.max.title.length", 100);

  private File directory = null;
  private int maxDocs = Integer.MAX_VALUE;

  /** Determines the power of link analyis scores.  Each pages's boost is
   * set to <i>score<sup>scorePower</sup></i> where <i>score</i> is its link
   * analysis score and <i>scorePower</i> is the value passed to this method.
   */
  public void setScorePower(float power) { scorePower = power; }

  private void indexPages() throws Exception {
    IndexWriter writer
      = new IndexWriter(new File(directory, "index"),
                        new NutchDocumentAnalyzer(), true);
    writer.mergeFactor = 50;
    writer.infoStream = System.out;
    writer.setSimilarity(new NutchSimilarity());

    ArrayFile.Reader fetcher =
      new ArrayFile.Reader(new File(directory, FetcherOutput.DIR_NAME).toString());
    ArrayFile.Reader text =
      new ArrayFile.Reader(new File(directory,FetcherText.DIR_NAME).toString());

    int count = 0;
    try {
      String segmentName = directory.getCanonicalFile().getName();
      FetcherOutput fetcherOutput = new FetcherOutput();
      FetcherText fetcherText = new FetcherText();
      //create hashtable from pagerank text file
      Hashtable table = readFromFile();

      while (fetcher.next(fetcherOutput) != null && count++ < maxDocs) {
        text.next(fetcherText);

        if (!fetcherOutput.getSuccess())          // if the fetch failed
          continue;                   // don't index the page

        Document doc = makeDocument(segmentName, fetcher.key(),
                                    fetcherOutput, fetcherText , table);
        writer.addDocument(doc);
      }
    } catch (EOFException e) {
      LOG.warning("Unexpected EOF in: " + directory +
                  " at entry #" + count + ".  Ignoring.");
    } finally {
      fetcher.close();
      text.close();
    }
    System.out.println("Optimizing index...");
    writer.optimize();
    writer.close();
  }

  private Hashtable readFromFile() {
     String file = "/home/manacar/nutch/data/pr_result_table.txt";
//     String file = "C:/Documents and Settings/manacar.FUJI/jbproject/nuc/pr_result_table.txt";
     Hashtable table = new Hashtable();
     BufferedReader inp = null;
   System.out.println("PR result reading to hashtable..");
     try {
       inp = new BufferedReader(new FileReader(file));


     Library library = new Library();
  //   System.out.println( library.readLine(inp));
     int count = 0;

    while (!(library.eof(inp))) {
       String line1 = library.readLine(inp);
       String line2 = library.readLine(inp);
 //      System.out.println(line1);
       table.put(line1, line2);
       count++;
//       System.out.println(count);
     }
     inp.close();
     System.out.println("Reading buffer done ...");
     } catch(Exception e) {
      e.printStackTrace();
     }
     return table;
  }


  private Document makeDocument(String segmentName, long docNo,
                                FetcherOutput fetcherOutput,
                                FetcherText fetcherText, Hashtable table)
    throws Exception {

    FetchListEntry fle = fetcherOutput.getFetchListEntry();
    String url = fle.getPage().getURL().toString();
    String title = fetcherOutput.getTitle();

    if (title.length() > maxTitleLength) {        // truncate title if needed
      title = title.substring(0, maxTitleLength);
    }

    Document doc = new Document();

    // url is both stored and indexed, so it's both searchable and returned
    doc.add(Field.Text("url", url));

    // un-indexed fields: not searchable, but in hits and/or used by dedup
    doc.add(Field.UnIndexed("title", title));
    doc.add(Field.UnIndexed("digest", fetcherOutput.getMD5Hash().toString()));
    doc.add(Field.UnIndexed("docNo", Long.toString(docNo, 16)));
    doc.add(Field.UnIndexed("segment", segmentName));

    // content is indexed, so that it's searchable, but not stored in index
    doc.add(Field.UnStored("content", fetcherText.getText()));



  // anchors are indexed, so they're searchable, but not stored in index
    String[] anchors = fle.getAnchors();
    for (int i = 0; i < anchors.length; i++) {
      doc.add(Field.UnStored("anchor", anchors[i]));
    }

    // add title as anchor so it's searchable.  doesn't warrant its own field.
    doc.add(Field.UnStored("anchor", title));

    // compute boost
    float boost = (float)Math.pow(fle.getPage().getScore(), scorePower);
    // apply boost to all indexed fields
    doc.setBoost(boost);


    //here we add pagerank results as a string
    // store boost for use by explain and dedup
    String pagerank_result_set = (String) table.get(url);
    if (pagerank_result_set == null) {
      pagerank_result_set = "-1";
    }
 //   String pr="";
 //   for (int i = 0; i < 1000; i++) {
 //     pr = pr + "1233323." + String.valueOf(i);
 //   }
    doc.add(Field.UnIndexed("boost", pagerank_result_set));

    return doc;
  }


  /** Create an index for the input files in the named directory. */
  public static void main(String[] args) throws Exception {

    String usage = "IndexSegment <segment_directory>";

    if (args.length == 0) {
      System.err.println("Usage: " + usage);
      return;
    }

    IndexSegment indexer = new IndexSegment();

    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-max")) {        // parse -max option
        indexer.maxDocs = Integer.parseInt(args[++i]);
      } else if (i != args.length-1) {
        System.err.println("Usage: " + usage);
        return;
      } else {
        indexer.directory = new File(args[i]);
      }
    }

    File fetcherDone = new File(indexer.directory, FetcherOutput.DONE_NAME);
    if (!fetcherDone.exists())                    // check fetcher done file
      throw new RuntimeException("can't index--not yet fetched: " +
                                 fetcherDone + " does not exist");

    File doneFile = new File(indexer.directory, DONE_NAME);
    if (doneFile.exists())                        // check index done file
      throw new RuntimeException("already indexed: " + doneFile + " exists");

    Date start = new Date();

    indexer.indexPages();

    Date end = new Date();

    System.out.print(end.getTime() - start.getTime());
    System.out.println(" total milliseconds");

    doneFile.createNewFile();                     // create the done file
  }

}
