/* Copyright (c) 2003 The Nutch Organization.  All rights reserved.   */
/* Use subject to the conditions in http://www.nutch.org/LICENSE.txt. */

package net.nutch.util;

import java.io.*;

/**
 * A collection of file-processing util methods
 */
public class FileUtil {

    /**
     * Delete a directory and all its contents.  If
     * we throw an exception, the directory may be
     * partially-deleted.
     */
    public static void fullyDelete(File dir) throws IOException {
        File contents[] = dir.listFiles();
        for (int i = 0; i < contents.length; i++) {
            if (contents[i].isFile()) {
                if (! contents[i].delete()) {
                    throw new IOException("Could not delete " + contents[i].getPath());
                }
            } else {
                fullyDelete(contents[i]);
            }
        }
        if (! dir.delete()) {
            throw new IOException("Could not delete " + dir.getPath());
        }
    }

    /**
     * Copy a file's contents to a new location
     */
    public static void copyContents(File src, File dst) throws IOException {
        DataInputStream in = new DataInputStream(new FileInputStream(src));
        try {
            DataOutputStream out = new DataOutputStream(new FileOutputStream(dst));
            byte buf[] = new byte[2048];
            try {
                int readBytes = in.read(buf);

                while (readBytes >= 0) {
                    out.write(buf, 0, readBytes);
                    readBytes = in.read(buf);
                }
            } finally {
                out.close();
            }
        } finally {
            in.close();
        }
    }

    /**
     * Copy a file and/or directory and all its contents (whether
     * data or other files/dirs)
     */
    public static void recursiveCopy(File src, File dst) throws IOException {
        //
        // Resolve the real target.
        //
        if (dst.exists() && dst.isDirectory()) {
            dst = new File(dst, src.getName());
        } else if (dst.exists()) {
            throw new IOException("Destination " + dst + " already exists");
        }

        //
        // Copy the items
        //
        if (! src.isDirectory()) {
            //
            // If the source is a file, then just copy the contents
            //
            copyContents(src, dst);
        } else {
            //
            // If the source is a dir, then we need to copy all the subfiles.
            //
            dst.mkdirs();
            File contents[] = src.listFiles();
            for (int i = 0; i < contents.length; i++) {
                recursiveCopy(contents[i], new File(dst, contents[i].getName()));
            }
        }
    }
}
